/*
 * @(#)Status.cpp
 *
 * This file is part of webCDwriter - Network CD Writing.
 *
 * Copyright (C) 1999-2002 Jrg P. M. Haeger
 *
 * webCDwriter is free software. See CDWserver.cpp for details.
 */

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include "Config.h"
#include "HTTPRequest.h"
#include "HTTPResponse.h"
#include "HTTPServer.h"
#include "LineReader.h"
#include "PrintWriter.h"
#include "Server.h"
#include "ServerSocket.h"
#include "Session.h"
#include "Socket.h"
#include "Status.h"
#include "Store.h"
#include "String.h"
#include "Thread.h"
#include "Version.h"

template<class Item>
class RingBuffer
{
	Item *items;
	int size, n, head;
	Item init;
public:
	RingBuffer(int size, Item init)
	{
		items = new Item[size];
		this->size = size;
		n = head = 0;
		this->init = init;
	}
	Item add(Item item)
	{
		Item old = items[(head + n) % size];
		items[(head + n) % size] = item;
		if (n < size)
		{
			n++;
			old = init;
		}
		else
			head++;
		return old;
	}
	Item get(int i)
	{
		if (i >= n)
			throw "i >= n";
		return items[(head + i) % size];
	}
};

RingBuffer<char *> lastSessionResults(10, NULL);


class StatusServer: public Thread {
	void run() {
		log.debug("StatusServer", "started");

		ServerSocket serverSocket(config.getStatusPortNo());

		for (int i = 1; ; i++)
		try {
			Socket *socket = serverSocket.accept();
			log.debug("StatusServer",
				"Status-Request %d accepted", i);

			HTTPServer server(*socket);
			HTTPRequest request = server.nextRequest();
			String s = "http://";
			HTTPResponse response = HTTPResponse::movedPermanently(
				s + request.host + ":" + config.getPortNo());
			PrintWriter writer(socket->getOutputStream());
			response.write(writer);

			delete socket;
			log.debug("StatusServer",
				"Status-Request %d terminated", i);
		} catch (Error *e) {
			log.debug("StatusServer", "Error: %s", e->toString());
			delete e;
		} catch (Exception *e) {
			log.debug("StatusServer", "Exception: %s",
				e->getMessage().getBytes());
			delete e;
		} catch (const char *str) {
			log.debug("StatusServer", "Catched: %s", str);
		}
	}
};


Status::Status()
{
	log.putv("Status::Status()");
}

int Status::infoNo = 0;
int Status::sessionsNum0 = 0;
int Status::sessionsNumNew = 0;
int Status::MBwritten0 = 0;
int Status::MBwrittenNew = 0;

void Status::addResult(const char *date, const char *time,
	const char *user, int size, const char *result)
{
	if (strstr(result, "OK") != NULL)
	{
		sessionsNumNew++;
		MBwrittenNew += size / (1024 * 1024);
	}

	char str[1024], userStr[1024];
	if (config.getShowUsers())
		sprintf(userStr, "<td align=center>%s</td>", user);
	else
		userStr[0] = 0;
	sprintf(str, "<td>%s</td><td>%s</td>%s"
		"<td align=right>%d</td>"
		"<td>%s</td>",
		date, time, userStr, size, result);

	char *str2 = lastSessionResults.add(strDup(str));
	if (str2 != NULL)
		delete[] str2;
}

static int scanToken(const char *str, int &pos, int &length)
{
	while (str[pos] == ' '
		|| str[pos] == '\t')
		pos++;

	int i = pos;
	while (str[i] != 0
		&& str[i] != ' '
		&& str[i] != '\t'
		&& str[i] != '\n')
		i++;

	return length = i-pos;
}

void Status::readLastResults()
{
	char fileName[4*1024];
	sprintf(fileName, "%s/sessions", config.getLogDir());
	FILE *file = fopen(fileName, "r");
	if (file == NULL)
		return;

	while (1)
	{
		char lineStr[4*1024];
		if (fgets(lineStr, sizeof lineStr, file) == NULL)
			break;

		char date[80], time[80], user[80];
		int size;

		int pos = 0, len;
		scanToken(lineStr, pos, len);
		strncpy(date, &lineStr[pos], len);
		date[len] = 0;
		pos += len;

		scanToken(lineStr, pos, len);
		strncpy(time, &lineStr[pos], len);
		time[len] = 0;
		pos += len;

		scanToken(lineStr, pos, len);
		strncpy(user, &lineStr[pos], len);
		user[len] = 0;
		pos += len;

		scanToken(lineStr, pos, len); pos += len;
		scanToken(lineStr, pos, len); pos += len;

		scanToken(lineStr, pos, len);
		if (sscanf(&lineStr[pos], "%d", &size) != 1)
			continue;
		pos += len;

		scanToken(lineStr, pos, len); pos += len;

		scanToken(lineStr, pos, len);
		addResult(date, time, user, size, &lineStr[pos]);
	}

	fclose(file);

	sessionsNum0 = sessionsNumNew;
	sessionsNumNew = 0;

	MBwritten0 = MBwrittenNew;
	MBwrittenNew = 0;
}

void Status::start() {
	readLastResults();

	if (config.getStatusPortNo() > 0)
		(new StatusServer())->start();
}

String *Status::getStatus() {
	String str = "";
	str = str
		+ "<table bgcolor=\"#C8C8FF\" border=\"3\" cellpadding=\"3\">\n"
		+ "<tr><td>Version</td><td>" + version + "</td></tr>\n"
		+ "<tr><td>Author</td><td><a href=\"" + authorHome + "\">"
		+ author + "</a></td></tr>\n"
		+ "<tr><td>Uses</td><td>";

	if (strlen(config.getMpg123Info()) > 0)
		str = str
			+ " <a href=\"" + mpg123Home + "\">mpg123</a> "
			+ config.getMpg123Info();

	str = str
		+ " <a href=\"" + mkisofsHome + "\">"
		+ config.getMkisofsInfo() + "</a>"
		+ " <a href=\"" + cdrecordHome + "\">"
		+ config.getCdrecordInfo() + "</a>"
		+ "</td></tr>\n"
		+ "</table>\n";

/*
//	time_t t;
//	time(&t);
	str = str
		+ "<h3>Info " + ++infoNo + " / Status"
//		+ ctime(&t)
		+ " <!-- yyyy-mm-dd -->"
		+ " <!-- time -->"
		+ "</h3>\n";
*/
	str = str
		+ "<h3>Status Info " + ++infoNo + "</h3>";

	Error *firstError = config.getFirstError();
	if (firstError != NULL) {
		config.reload();
		firstError = config.getFirstError();
		if (firstError == NULL)
			Server::initStatic2();
	}

	if (firstError != NULL) {
		str = str
			+ "<h2>Server Configuration Error</h2>\n"
			+ "<h3>" + firstError->toString() + "</h3>\n";
		char str2[16 * 1024];
		strcpy(str2, firstError->getHelp());
		replace(str2, "[", "\n<table bgcolor=#E0E0E0 "
			"cellpadding=10 width=100%><tr><td>\n");
		replace(str2, "]", "\n</tr></td></table>\n");
		str = str
			+ str2 + "\n"
			+ "<h3>Reload this page to try again!<h3>\n";
		return new String(str);
	}

	if (store == NULL) {
		str = str
			+ "<h1>Problem: store == NULL</h1>\n"
			+ "<h2>Make sure you have "
			+ "glibc2/libc6 installed!</h2>\n";
		return new String(str);
	}

	str = str
		+ "<table bgcolor=\"#C8C8FF\" border=\"3\" cellpadding=\"3\">\n"
		+ "<tr><th></th><th></th>"
		+ "<th>Max</th><th></th></tr>\n"
		+ "<tr><td>Open connections</td>"
		+ "<td>" + Server::getActiveServersNum()
		+ "</td><td>" + config.getMaxOpenConnections() + "</td></tr>\n"
		+ "<tr><td>Open sessions</td>"
		+ "<td>" + store->getSessionsNum()
		+ "</td><td>" + config.getMaxOpenSessions() + "</td></tr>\n"
		+ "<tr><td>Cache</td>"
		+ "<td>" + store->getSizeInMB()
		+ " MB in " + store->getSessionsTotal()
		+ " sessions</td><td>" + config.getMaxMBytesPerSession()
		+ " / " + store->getMaxSizeInMB() + " MB</td>"
		+ "<td>session / total</td></tr>\n"
		+ "</table>\n";

	str = str
		+ "<table cellspacing=\"10\" width=\"100%\"><tr>\n";

	if (config.getCopySupport()) {
		str = str
			+ "<td>\n"
			+ "<h3>CD-ROM(s)</h3>\n"
			+ "<table bgcolor=\"#C8C8FF\" border=\"3\" cellpadding=\"3\">\n"
			+ "<tr><th>No</th><th>Type</th>"
			+ "<th>Status</th></tr>\n";

		for (int i = 0; i < config.getNumOfReaders(); i++) {
			const char *status = "";
			switch (Server::getReaderStatus(i)) {
				case idle:
					status = "free";
					break;
				case reading:
					status = "reading";
					break;
				case reserved:
					status = "reserved";
					break;
			}

			str = str
				+ "<tr><td>" + (i + 1)
				+ "</td><td>" + config.getReaderInfo(i)
				+ "</td><td>"
				+ status + "</td></tr>";
		}
		str = str
			+ "</table>\n"
			+ "</td>\n";
	}

	str = str
		+ "<td>\n"
		+ "<h3>CD-writer(s)</h3>\n"
		+ "<table bgcolor=\"#C8C8FF\" border=\"3\" cellpadding=\"3\">\n"
		+ "<tr><th>No</th><th>Type</th>"
		+ "<th>Status</th></tr>\n";

	for (int i = 0; i < config.getNumOfWriters(); i++) {
		str = str
			+ "<tr><td>" + (i + 1)
			+ "</td><td>" + config.getWriterInfo(i) + "</td><td>";

		const char *status = NULL;
		switch (Server::getWriterStatus(i)) {
			case idle:
				str = str + "free";
				break;
			case reserved: {
				str = str + "reserved";
				if (config.getShowUsers()) {
					Server *s = Server::getWriterOwner(i);
					str = str
						+ " by " + s->getClientUserName()
						+ "@" + s->getClientIP();
				}
				struct timeval t;
				gettimeofday(&t, NULL);
				t.tv_sec = min(t.tv_sec,
					Server::getWriterReservedUntil(i));
				char str2[1024];
				str = str
					+ "</td><td>"
					+ secondsToString(
						Server::getWriterReservedUntil(i)
							- t.tv_sec, str2);
				break;
			}
			case blanking:
				status = "blanking CD-RW (cdrecord)";
				break;
			case mkisofs:
				status = "creating image (mkisofs)";
				break;
			case writing:
				status = "writing (cdrecord)";
				break;
			case fixating:
				status = "fixating (cdrecord)";
				break;
			case verifying:
				status = "verifying";
				break;
		}

		if (status != NULL)
			str = str
				+ status + " "
				+ Server::getWriterPercent(i) + "%";

		str = str + "</td></tr>";
	}
	str = str + "</table>\n" + "</td></tr></table>\n";

	str = str
		+ "<h3>Last sessions written</h3>\n"
		+ "<table border=\"3\" cellpadding=\"3\">\n"
		+ "<tr bgcolor=\"#C8C8FF\">\n"
		+ "<th>Date</th><th>Time</th>";
	if (config.getShowUsers())
		str = str + "<th>User</th>";
	str = str
		+ "<th>Size</th><th>Result</th>\n</tr>\n";

	try {
		int i = 0;
		while (1)
			str = str
				+ "<tr>" + lastSessionResults.get(i++) + "</tr>\n";
	} catch (const char *msg) {
	}
	str = str + "</table>\n";

	str = str
		+ "<h3>Statistics</h3>\n"
		+ "<table bgcolor=\"#C8C8FF\" border=\"3\" cellpadding=\"3\">\n"
		+ "<tr><td></td><th>new</th><th>total</th></tr>\n"
		+ "<tr><td>Sessions</td>"
		+ "<td>" + sessionsNumNew +
		+ "</td><td>" + (sessionsNum0 + sessionsNumNew) + "</td></tr>\n"
		+ "<tr><td>GB written</td>"
		+ "<td>" + (MBwrittenNew / 1024)
		+ "</td><td>" + ((MBwritten0 + MBwrittenNew) / 1024)
		+ "</td></tr>\n"
		+ "</table>\n";

	return new String(str);
}

Status status;
